#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2021-08-20

# based on the BGP peer plugin by Thomas Wollner (tw@wollner-net.de)
#
# 2021-08-20: rewritten for CMK 2.0 by thl-cmk[at]outlook[dot]com
# 2021-08-21: added more perfdata, added metrics file
# 2021-08-22: added WATO options, streamlined with cisco_bgp_peer
# 2021-08-29: moved static (longoutput) to inventory plugin
#             moved helper functions to utils/bgp_peer
# 2021-11-14: merged check function with cisco_bgp_peer
#             moved parse function to utils/bgp_peer
# 2021-04-02: rewritten bgp neighbor state handling (made configurable)
# 2022-04-29: added upper/lower prefix limits from wato
#             added info if device is admin prefix limit capable (device_admin_limit)
# 2022-05-09: made item name configurable (don't use address-family/routing-instance/VRF)
# 2022-05-11: changed bgp_get_peer_entry to get proper parameters instead of Nontransparent list
#             added remote_as to BgpPeerItem
# 2022-09-05: added missing wato parameters to register.check_plugin check_default_parameters
# 2022-09-10: made more reliable on limited data (fsm_established_time and admin_state)
#             (THX to martin[dot]pechstein[at]posteo[dot]de)
# 2022-09-11: optimized internal flow: > alias > not found > admin down > peer state > ...
# 2023-01-21: changed to always yield fsm_established_time (not only if beep connects, but also on all other states)
# 2023-01-22: fix output for admin_state
# 2023-02-16: changed for CMK 2.1 (moved gui files from local/share/.. to local/lib/..)
#             fix type error in discovery (CMK2.1 GUI only)
# 2023-02-20: fix crash if metric data is None
# 2023-03-26: optimized output of metrics, GetRateError will not be set to 0 anymore
# 2023-08-17: fix removed internal_item form check_default_parameters (THX mail[at]bastian-kuhn[dot]de)
#             added warning if internal_item is missing in params
# 2024-06-07: fixed crash on configured state mapping
# 2024-12-20: added local address/as/identifier for bgp_topology
# 2025-03-29: refactored for ruleset APIv1

# Example Agent Output:
# BGP4-MIB

# .1.3.6.1.2.1.15.3.1.1.192.168.254.2 = IpAddress: 192.168.254.2
# .1.3.6.1.2.1.15.3.1.2.192.168.254.2 = INTEGER: 3
# .1.3.6.1.2.1.15.3.1.3.192.168.254.2 = INTEGER: 2
# .1.3.6.1.2.1.15.3.1.4.192.168.254.2 = INTEGER: 4
# .1.3.6.1.2.1.15.3.1.5.192.168.254.2 = IpAddress: 0.0.0.0
# .1.3.6.1.2.1.15.3.1.7.192.168.254.2 = IpAddress: 192.168.254.2
# .1.3.6.1.2.1.15.3.1.9.192.168.254.2 = INTEGER: 65301
# .1.3.6.1.2.1.15.3.1.10.192.168.254.2 = Counter32: 0
# .1.3.6.1.2.1.15.3.1.11.192.168.254.2 = Counter32: 5
# .1.3.6.1.2.1.15.3.1.14.192.168.254.2 = Hex-STRING: 04 00
# .1.3.6.1.2.1.15.3.1.16.192.168.254.2 = Gauge32: 6586822
#

import time
from typing import Dict, List, Generator

from cmk.agent_based.v2 import (
    CheckPlugin,
    CheckResult,
    DiscoveryResult,
    GetRateError,
    Metric,
    Result,
    SNMPSection,
    SNMPTree,
    Service,
    State,
    StringTable,
    check_levels,
    exists,
    get_rate,
    get_value_store,
    render,
)

from cmk_addons.plugins.bgp_peer.lib.bgp_peer import (
    BgpPeer,
    bgp_adminstate,
    bgp_get_peer_entry,
    bgp_peerstate,
)


def parse_bgp_peer(string_table: List[StringTable]) -> Dict[str, BgpPeer] | None:
    peer_table = {}
    try:
        peer_info, local_info = string_table
    except ValueError:
        return None

    try:
        local_as, local_id = local_info[0]
    except (IndexError, ValueError):
        local_as, local_id = None, None

    for entry in peer_info:
        remote_addr, remote_id, peer_state, admin_state, local_addr, remote_as, in_updates, out_updates, \
            in_messages, out_messages, fsm_established_transitions, fsm_established_time, in_update_elapsed_time = entry

        bgp_peer = bgp_get_peer_entry(
            admin_state=admin_state,
            fsm_established_time=fsm_established_time,
            fsm_established_transitions=fsm_established_transitions,
            in_messages=in_messages,
            in_update_elapsed_time=in_update_elapsed_time,
            in_updates=in_updates,
            local_addr=local_addr,
            out_messages=out_messages,
            out_updates=out_updates,
            peer_state=peer_state,
            remote_addr=remote_addr,
            remote_as=remote_as,
            remote_id=remote_id,
            local_as=local_as,
            local_id=local_id,
        )
        if bgp_peer:
            peer_table.update(bgp_peer)
    return peer_table


def discovery_bgp_peer(params, section: Dict[str, BgpPeer]) -> DiscoveryResult:
    _item_parts = [
        'remote_address',
        'remote_as',
        'address_family',
        'routing_instance',
    ]
    for key in section.keys():
        parameters = {'internal_item': key}
        item = ''
        for item_part in _item_parts:
            if item_part not in params['build_item']:
                item += f'{section[key].item[item_part]} '
        item = item.strip(' ')
        yield Service(item=item, parameters=parameters)


def bgp_peer_base_info(peer: BgpPeer) -> Generator:
    for message, value in [
        ('Local AS:', peer.local_as),
        ('Local address:', peer.local_addr),
        ('Local identifier:', peer.local_id),
        ('Remote AS:', peer.remote_as),
        ('Remote address:', peer.remote_addr),
        ('Remote identifier:', peer.remote_id),
    ]:
        if value is not None:
            yield Result(state=State.OK, notice=f'{message} {value}')


def _rewrite_ruleset_needed(params: dict) -> bool:
    for param, value in params.items():
        match param:
            case 'accepted_prefixes_lower_levels' | 'accepted_prefixes_upper_levels' | 'minuptime':
                if isinstance(value, tuple) and len(value) == 2 \
                        and isinstance(value[0], int) and isinstance(value[1], int):
                    return True
            case 'neighborstate':
                for key in value.keys():
                    if key in '123456':
                        return True
            case 'peer_list':
                if isinstance(value[0], tuple):
                    return True
    return False


def check_bgp_peer(item, params, section: Dict[str, BgpPeer]) -> CheckResult:
    if not params.get('internal_item'):
        yield Result(
            state=State.WARN,
            summary='This service is using old parameters (see details)',
            details='This service is using old parameters. To refresh the parameters you need to do a "Tabula rasa" '
                    '(Discover services -> Remove all and find new)',
        )
    else:
        item = params['internal_item']

    if _rewrite_ruleset_needed(params):
        yield Result(state=State.WARN,
                     summary='BGP ruleset needs to be opened and saved to migrate the ruleset to ruleset APIv1!')
        return

    neighbor_state = {
        '1': 2,  # idle
        '2': 1,  # connect
        '3': 1,  # active
        '4': 1,  # opensent
        '5': 1,  # openconfirm
        '6': 0,  # established
    }

    neighbor_state.update(params['neighborstate'])
    peer_not_found_state = params['peernotfound']

    for entry in params.get('peer_list', []):
        if item == entry['bgp_peer']:
            if entry.get('bgp_peer_alias'):
                yield Result(state=State.OK, summary=f'[{entry["bgp_peer_alias"]}]', details=' ')
            peer_not_found_state = entry.get('state_not_found', params['peernotfound'])

    # get item data
    try:
        peer = section[item]
    except KeyError:
        yield Result(state=State(peer_not_found_state), summary='Item not found in SNMP data')
        return

    if peer.description:
        yield Result(state=State.OK, summary=f'[{peer.description}]', details=' ')

    if peer.admin_state == 1:  # shutdown
        yield Result(state=State(params['admindown']), summary=f'Admin state: {bgp_adminstate(peer.admin_state)}')
        yield from bgp_peer_base_info(peer)
        return

    yield Result(
        state=State(neighbor_state.get(str(peer.peer_state))),
        notice=f'Peer state: {bgp_peerstate(peer.peer_state)}'
    )

    if peer.fsm_established_time:  # always yield fsm_established time
        yield Metric(
            name=f'bgp_peer_fsmestablishedtime',
            value=peer.fsm_established_time,
            boundaries=(0, None)
        )

    if not peer.peer_state == 6:  # not established
        yield from bgp_peer_base_info(peer)
        return

    if peer.fsm_established_time:  # if peer is established use fsm_established_time as uptime
        yield from check_levels(
            value=peer.fsm_established_time,
            label='Uptime',
            levels_upper=('no_levels', None),
            levels_lower=params['minuptime'],
            render_func=render.timespan,
            metric_name='bgp_peer_fsmestablishedtime',
        )

    if peer.peer_unavail_reason != 0:  # huawei peer unavailable state
        yield Result(state=State.CRIT, notice=F'Peer unavailable reason: {peer.peer_unavail_reason_str}')

    if peer.device_admin_limit and peer.prefix_admin_limit is None:
        yield Result(
            state=State(params['noprefixlimit']),
            notice='Prefix limit/warn threshold not configured on the device.',
        )

    if peer.accepted_prefixes is not None and peer.peer_state == 6:  # peer established and prefixes accepted
        if peer.prefix_admin_limit is None or peer.prefix_threshold is None:
            levels_upper = None
        else:
            # use float (100.0) to get xx.xx in division
            levels_upper = (
                'fixed', (int(peer.prefix_admin_limit / 100.0 * peer.prefix_threshold), peer.prefix_admin_limit)
            )

        yield from check_levels(
            value=peer.accepted_prefixes,
            metric_name='bgp_peer_acceptedprefixes',
            levels_upper=params.get('accepted_prefixes_upper_levels', levels_upper),
            levels_lower=params.get('accepted_prefixes_lower_levels'),
            label='Prefixes accepted',
            render_func=lambda v: f'{v}'
        )

    now_time = time.time()
    value_store = get_value_store()

    for key, value in peer.metric_rate:
        try:
            value = get_rate(value_store, f'{key}', now_time, value, raise_overflow=True)
        except GetRateError:
            continue
        yield Metric(name=f'bgp_peer_{key}', value=value, boundaries=(0, None))

    for key, value in peer.metric_count:
        if value is not None:
            yield Metric(name=f'bgp_peer_{key}', value=value, boundaries=(0, None))

    yield from bgp_peer_base_info(peer)


snmp_section_bgp_peer = SNMPSection(
    name='bgp_peer',
    parse_function=parse_bgp_peer,
    fetch=[
        SNMPTree(
            base='.1.3.6.1.2.1.15.3.1',  # BGP4-MIB::BgpPeerEntry
            oids=[
                '7',  # bgpPeerRemoteAddr
                '1',  # bgpPeerIdentifier
                '2',  # bgpPeerState
                '3',  # bgpPeerAdminStatus
                '5',  # bgpPeerLocalAddr
                '9',  # bgpPeerRemoteAs
                '10',  # bgpPeerInUpdates
                '11',  # bgpPeerOutUpdates
                '12',  # bgpPeerInTotalMessages
                '13',  # bgpPeerOutTotalMessages
                '15',  # bgpPeerFsmEstablishedTransitions
                '16',  # bgpPeerFsmEstablishedTime
                '24',  # bgpPeerInUpdateElapsedTime
            ]),
        SNMPTree(
            base='.1.3.6.1.2.1.15',  # BGP4-MIB
            oids=[
                '2',  # bgpLocalAs
                '4',  # bgpIdentifier (local)
            ])
    ],
    detect=exists('.1.3.6.1.2.1.15.3.1.1.*')
)

check_plugin_bgp_peer = CheckPlugin(
    name='bgp_peer',
    service_name='BGP peer %s',
    discovery_function=discovery_bgp_peer,
    discovery_default_parameters={
        'build_item': [
            # 'remote_address',
            'remote_as',
            # 'address_family',
            # 'routing_instance',
        ]
    },
    discovery_ruleset_name='discovery_bgp_peer',
    check_function=check_bgp_peer,
    check_default_parameters={
        'minuptime': ('fixed', (7200, 3600)),
        'peer_list': [],
        'peernotfound': 2,
        'admindown': 1,
        'noprefixlimit': 1,
        'neighborstate': {
            'idle': 2,  # ........1
            'connect': 1,  # .....2
            'active': 1,  # ......3
            'open_sent': 1,  # ...4
            'open_confirm': 1,  # 5
            'established': 0,  # .6
        },
    },
    check_ruleset_name='bgp_peer',
)
