#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2022-04-24
#
# 2022-04-24: added option for BGP down time
#             added option to remove some columns from inventory
# 2022-04-28: added Whois options
# 2023-06-12: moved wato file from ~local/lib/check_mk/gui/plugins/wato
#             to ~/local/lib/check_mk/gui/plugins/wato/check_parameters
# 2025-03-29: rewritten for rulesets API v1

from cmk.rulesets.v1 import Title, Help
from cmk.rulesets.v1.form_specs import (
    DefaultValue,
    DictElement,
    Dictionary,
    Integer,
    MultipleChoice,
    MultipleChoiceElement,
    SingleChoice,
    SingleChoiceElement,
    TimeMagnitude,
    TimeSpan,
)
from cmk.rulesets.v1.rule_specs import InventoryParameters, Topic
from cmk.rulesets.v1.form_specs.validators import NumberInRange, Message

def _migrate_to_float(value: object) -> float:
    if isinstance(value, int | float):
        return float(value)

    raise TypeError(value)


def _parameter_form_inv_bgp_peer():
    remove_columns = [
        # MultipleChoiceElement(name='remote_as', title=Title('Remote AS')),
        # MultipleChoiceElement(name='remote_id', title=Title('Remote ID')),
        # MultipleChoiceElement(name='local_addr', title=Title('Local address')),
        # MultipleChoiceElement(name='local_as', title=Title('Local AS')),
        # MultipleChoiceElement(name='local_id', title=Title('Local ID')),
        # MultipleChoiceElement(name='prev_state', title=Title('Previous state')),
        MultipleChoiceElement(name='address_family', title=Title('Address family')),
        MultipleChoiceElement(name='last_error', title=Title('Last error')),
        MultipleChoiceElement(name='last_error_code', title=Title('Last error code')),
        MultipleChoiceElement(name='as_name', title=Title('Remote AS name')),
        MultipleChoiceElement(name='as_org_name', title=Title('Remote AS Org Name')),
        MultipleChoiceElement(name='bgp_type', title=Title('Type')),
        MultipleChoiceElement(name='version', title=Title('Version')),
    ]

    rir_s = [
        SingleChoiceElement(name='afrinic', title=Title('AFRINIC (https://rdap.afrinic.net/rdap)')),
        SingleChoiceElement(name='apnic', title=Title('APNIC (https://rdap.apnic.net)')),
        SingleChoiceElement(name='arin', title=Title('ARIN (https://rdap.arin.net/registry)')),
        SingleChoiceElement(name='ripe', title=Title('RIPE (https://rdap.db.ripe.net)')),
        SingleChoiceElement(name='lacnic', title=Title('LACNIC (https://rdap.apnic.net)')),
    ]

    return Dictionary(
        elements={
            'not_in_service_time': DictElement(
                parameter_form=TimeSpan(
                    title=Title('Time peer is not up until considered not in service'),
                    displayed_magnitudes=[TimeMagnitude.MINUTE, TimeMagnitude.HOUR, TimeMagnitude.DAY],
                    prefill=DefaultValue(2592000.0),  # 30 days in seconds
                    migrate=_migrate_to_float,
                )),
            'remove_columns': DictElement(
                parameter_form=MultipleChoice(
                    title=Title('List of columns to remove'),
                    help_text=Help('Information to remove from inventory'),
                    elements=remove_columns,
                )),
            'whois_enable': DictElement(
                parameter_form=Dictionary(
                    title=Title('Add whois data to the inventory'),
                    help_text=Help(
                        'The whois data will be fetched via RDAP from the registries. For this the the plugin tries to'
                        'find the best registry via the RDAP bootstrap data from https://data.iana.org/rdap/asn.json.'
                        'The query it self will go to the found registry via http(s). Note: the request might be get '
                        'redirected if there a different authoritative registry for the ASn'
                    ),
                    elements={
                        'whois_rir': DictElement(
                            parameter_form=SingleChoice(
                                title=Title('Preferred RIR to fetch whois data'),
                                help_text=Help(
                                    'This registry will be used if the plugin can not determine the authoritative registry '
                                    'based on the bootstrap data.'
                                ),
                                elements=rir_s,
                                prefill=DefaultValue('ripe'),
                            )),
                        'whois_timeout': DictElement(
                            parameter_form=Integer(
                                custom_validate=(
                                    NumberInRange(
                                        min_value=1,
                                        max_value=None,
                                        error_msg=Message('The timeout must be greater than 0')
                                    ),
                                ),
                                help_text=Help('The connection timeout for each whois request.'),
                                prefill=DefaultValue(5),
                                title=Title('Timeout for connections to RIRs'),
                                unit_symbol='seconds',
                            )),
                    }
                )
            )
        }
    )


rule_spec_inv_bgp_peer = InventoryParameters(
    name='inv_bgp_peer',
    parameter_form=_parameter_form_inv_bgp_peer,
    title=Title('BGP peer'),
    topic=Topic.NETWORKING,
)
